@@ -26,6 +26,7 @@
 
 #include "config.h"
 
+#include "libwacom.h"
 #include "libwacomint.h"
 #include <assert.h>
 #include <stdlib.h>
@@ -212,11 +213,72 @@ client_query_by_subsystem_and_device_file (GUdevClient *client,
 	return ret;
 }
 
+static char *
+get_device_prop(GUdevDevice *device, const char *propname)
+{
+	char *value = NULL;
+	GUdevDevice *parent = g_object_ref(device);
+
+	do {
+		GUdevDevice *next;
+		const char *v = g_udev_device_get_property(parent, propname);
+		if (v) {
+			/* NAME and UNIQ properties are enclosed with quotes */
+			size_t offset = v[0] == '"' ? 1 : 0;
+			value = g_strdup(v + offset);
+			if (value[strlen(value)] == '"')
+				value[strlen(value)] = '\0';
+			break;
+		}
+		next = g_udev_device_get_parent (parent);
+		g_object_unref(parent);
+		parent = next;
+	} while (parent);
+
+	if (parent)
+		g_object_unref(parent);
+
+	return value;
+}
+
+static char *
+parse_uniq(char *uniq)
+{
+	GRegex *regex;
+	GMatchInfo *match_info;
+
+	if (!uniq)
+		return NULL;
+
+	if (strlen(uniq) == 0) {
+		g_free (uniq);
+		return NULL;
+	}
+
+	/* The UCLogic kernel driver returns firmware names with form
+	 * <vendor>_<model>_<version>. Remove the version from `uniq` to avoid
+	 * mismatches on firmware updates. */
+	regex = g_regex_new ("(.*_.*)_.*$", 0, 0, NULL);
+	g_regex_match (regex, uniq, 0, &match_info);
+
+	if (g_match_info_matches (match_info)) {
+		gchar *tmp = uniq;
+		uniq = g_match_info_fetch (match_info, 1);
+		g_free (tmp);
+	}
+
+	g_match_info_free (match_info);
+	g_regex_unref (regex);
+
+	return uniq;
+}
+
 static gboolean
 get_device_info (const char            *path,
 		 int                   *vendor_id,
 		 int                   *product_id,
 		 char                 **name,
+		 char                 **uniq,
 		 WacomBusType          *bus,
 		 WacomIntegrationFlags *integration_flags,
 		 WacomError            *error)
@@ -232,6 +294,7 @@ get_device_info (const char            *path,
 	/* The integration flags from device info are unset by default */
 	*integration_flags = WACOM_DEVICE_INTEGRATED_UNSET;
 	*name = NULL;
+	*uniq = NULL;
 	bus_str = NULL;
 	client = g_udev_client_new (subsystems);
 	device = client_query_by_subsystem_and_device_file (client, subsystems[0], path);
@@ -283,17 +346,10 @@ get_device_info (const char            *path,
 		g_free (sysfs_path);
 	}
 
-	*name = g_strdup (g_udev_device_get_sysfs_attr (device, "name"));
-	/* Try getting the name from the parent if that fails */
-	if (*name == NULL) {
-		GUdevDevice *parent;
-
-		parent = g_udev_device_get_parent (device);
-		if (!parent)
-			goto out;
-		*name = g_strdup (g_udev_device_get_sysfs_attr (parent, "name"));
-		g_object_unref (parent);
-	}
+	*name = get_device_prop (device, "NAME");
+	*uniq = parse_uniq(get_device_prop(device, "UNIQ"));
+	if (*name == NULL)
+		goto out;
 
 	/* Parse the PRODUCT attribute (for Bluetooth, USB, I2C) */
 	retval = get_bus_vid_pid (device, bus, vendor_id, product_id, error);
@@ -318,8 +374,10 @@ get_device_info (const char            *path,
 out:
 	if (bus_str != NULL)
 		g_free (bus_str);
-	if (retval == FALSE)
+	if (retval == FALSE) {
 		g_free (*name);
+		g_free (*uniq);
+	}
 	if (device != NULL)
 		g_object_unref (device);
 	if (client != NULL)
@@ -354,8 +412,12 @@ libwacom_copy(const WacomDevice *device)
 		d->paired = libwacom_match_ref(device->paired);
 	d->cls = device->cls;
 	d->num_strips = device->num_strips;
+	d->num_rings = device->num_rings;
+	d->num_dials = device->num_dials;
 	d->features = device->features;
 	d->strips_num_modes = device->strips_num_modes;
+	d->dial_num_modes = device->dial_num_modes;
+	d->dial2_num_modes = device->dial2_num_modes;
 	d->ring_num_modes = device->ring_num_modes;
 	d->ring2_num_modes = device->ring2_num_modes;
 	d->styli = g_array_sized_new(FALSE, FALSE, sizeof(int),
@@ -472,12 +534,21 @@ libwacom_compare(const WacomDevice *a, const WacomDevice *b, WacomCompareFlags f
 	if (a->num_strips != b->num_strips)
 		return 1;
 
+	if (a->num_dials != b->num_dials)
+		return 1;
+
 	if (a->features != b->features)
 		return 1;
 
 	if (a->strips_num_modes != b->strips_num_modes)
 		return 1;
 
+	if (a->dial_num_modes != b->dial_num_modes)
+		return 1;
+
+	if (a->dial2_num_modes != b->dial2_num_modes)
+		return 1;
+
 	if (a->ring_num_modes != b->ring_num_modes)
 		return 1;
 
@@ -524,7 +595,7 @@ libwacom_compare(const WacomDevice *a, const WacomDevice *b, WacomCompareFlags f
 }
 
 static const WacomDevice *
-libwacom_new (const WacomDeviceDatabase *db, const char *name, int vendor_id, int product_id, WacomBusType bus, WacomError *error)
+libwacom_new (const WacomDeviceDatabase *db, const char *name, const char *uniq, int vendor_id, int product_id, WacomBusType bus, WacomError *error)
 {
 	const WacomDevice *device;
 	char *match;
@@ -534,24 +605,94 @@ libwacom_new (const WacomDeviceDatabase *db, const char *name, int vendor_id, in
 		return NULL;
 	}
 
-	match = make_match_string(name, bus, vendor_id, product_id);
+	match = make_match_string(name, uniq, bus, vendor_id, product_id);
 	device = libwacom_get_device(db, match);
 	g_free (match);
 
 	return device;
 }
 
-LIBWACOM_EXPORT WacomDevice*
-libwacom_new_from_path(const WacomDeviceDatabase *db, const char *path, WacomFallbackFlags fallback, WacomError *error)
+static bool
+builder_is_name_only(const WacomBuilder *builder)
 {
-	int vendor_id, product_id;
-	WacomBusType bus;
-	const WacomDevice *device;
-	WacomDevice *ret = NULL;
-	WacomIntegrationFlags integration_flags;
-	char *name, *match_name;
-	WacomMatch *match;
+	return builder->device_name != NULL && builder->match_name == NULL &&
+		builder->uniq == NULL &&
+		builder->vendor_id == 0 && builder->product_id == 0 &&
+		builder->bus == WBUSTYPE_UNKNOWN;
+}
 
+static bool
+builder_is_uniq_only(const WacomBuilder *builder)
+{
+	return builder->device_name == NULL && builder->match_name == NULL &&
+		builder->uniq != NULL &&
+		builder->vendor_id == 0 && builder->product_id == 0 &&
+		builder->bus == WBUSTYPE_UNKNOWN;
+}
+
+/**
+ * Return a copy of the given device or, if NULL, the fallback device with
+ * the name changed to the override name.
+ */
+static WacomDevice *
+fallback_or_device(const WacomDeviceDatabase *db, const WacomDevice *device,
+		   const char *name_override, WacomFallbackFlags fallback_flags)
+{
+	WacomDevice *copy = NULL;
+	const WacomDevice *fallback;
+	const char *fallback_name = NULL;
+
+	if (device != NULL) {
+		return libwacom_copy(device);
+	}
+
+	switch (fallback_flags) {
+	case WFALLBACK_NONE:
+		return NULL;
+	case WFALLBACK_GENERIC:
+		fallback_name = "generic";
+		break;
+	default:
+		g_assert_not_reached();
+		break;
+	}
+
+	fallback = libwacom_get_device(db, fallback_name);
+	if (fallback == NULL)
+		return NULL;
+
+	copy = libwacom_copy(fallback);
+	if (name_override != NULL) {
+		g_free(copy->name);
+		copy->name = g_strdup(name_override);
+	}
+	return copy;
+}
+
+static gint
+find_named_device(const WacomDevice *device, const char *name)
+{
+	return g_strcmp0(device->name, name);
+}
+
+static gint
+find_uniq_device(const WacomDevice *device, const char *uniq)
+{
+	const WacomMatch **matches = libwacom_get_matches(device);
+
+	for (const WacomMatch **match = matches; *match; match++) {
+		if (uniq && (*match)->uniq && g_str_equal((*match)->uniq, uniq))
+			return 0;
+	}
+	return -1;
+}
+
+LIBWACOM_EXPORT WacomDevice*
+libwacom_new_from_builder(const WacomDeviceDatabase *db, const WacomBuilder *builder,
+			  WacomFallbackFlags fallback, WacomError *error)
+{
+	const WacomDevice *device = NULL;
+	WacomDevice *ret = NULL;
 	switch (fallback) {
 		case WFALLBACK_NONE:
 		case WFALLBACK_GENERIC:
@@ -566,109 +707,157 @@ libwacom_new_from_path(const WacomDeviceDatabase *db, const char *path, WacomFal
 		return NULL;
 	}
 
-	if (!path) {
-		libwacom_error_set(error, WERROR_INVALID_PATH, "path is NULL");
-		return NULL;
-	}
-
-	if (!get_device_info (path, &vendor_id, &product_id, &name, &bus, &integration_flags, error))
-		return NULL;
-
-	match_name = name;
-	device = libwacom_new (db, match_name, vendor_id, product_id, bus, error);
-	if (device == NULL) {
-		match_name = NULL;
-		device = libwacom_new (db, match_name, vendor_id, product_id, bus, error);
-	}
-
-	if (device == NULL) {
-		if (fallback == WFALLBACK_NONE)
-			goto out;
-
-		/* WFALLBACK_GENERIC */
-		device = libwacom_get_device(db, "generic");
-		if (device == NULL)
-			goto out;
-
-		ret = libwacom_copy(device);
-
-		if (name != NULL) {
-			g_free (ret->name);
-			ret->name = g_strdup(name);
-		}
+	/* Name-only matches behave like new_from_name */
+	if (builder_is_name_only(builder)) {
+		GList *keys = g_hash_table_get_values(db->device_ht);
+		GList *entry = g_list_find_custom(keys, builder->device_name, (GCompareFunc)find_named_device);
+		if (entry)
+			device = entry->data;
+		ret = fallback_or_device(db, device, builder->device_name, fallback);
+		g_list_free (keys);
+	/* Uniq-only behaves like new_from_name but matches on uniq in the match strings */
+	} else if (builder_is_uniq_only(builder)) {
+		GList *keys = g_hash_table_get_values(db->device_ht);
+		GList *entry = g_list_find_custom(keys, builder->uniq, (GCompareFunc)find_uniq_device);
+		if (entry)
+			device = entry->data;
+		ret = fallback_or_device(db, device, builder->device_name, fallback);
+		g_list_free (keys);
 	} else {
-		ret = libwacom_copy(device);
+		WacomBusType all_busses[] = {
+			WBUSTYPE_USB,
+			WBUSTYPE_I2C,
+			WBUSTYPE_BLUETOOTH,
+			WBUSTYPE_UNKNOWN,
+		};
+		WacomBusType fixed_bus[] = {
+			builder->bus,
+			WBUSTYPE_UNKNOWN,
+		};
+		WacomBusType *bus;
+
+		int vendor_id, product_id;
+		char *name, *uniq;
+		const char *used_match_name = NULL;
+		const char *used_match_uniq = NULL;
+		WacomMatch *used_match;
+
+		vendor_id = builder->vendor_id;
+		product_id = builder->product_id;
+		name = builder->match_name;
+		uniq = builder->uniq;
+		if (builder->bus)
+			bus = fixed_bus;
+		else
+			bus = all_busses;
+
+		while (*bus != WBUSTYPE_UNKNOWN) {
+			/* Uniq (where it exists) is more reliable than the name which may be re-used
+			 * across tablets. So try to find a uniq+name match first, then uniq-only, then
+			 * name-only.
+			 */
+			struct match_approach {
+				const char *name;
+				const char *uniq;
+			} approaches[] = {
+				{ name, uniq },
+				{ NULL, uniq },
+				{ name, NULL },
+				{ NULL, NULL },
+			};
+			struct match_approach *approach = approaches;
+			while (true) {
+				const char *match_name = approach->name;
+				const char *match_uniq = approach->uniq;
+				device = libwacom_new (db, match_name, match_uniq, vendor_id, product_id, *bus, error);
+				if (device) {
+					used_match_name = match_name;
+					used_match_uniq = match_uniq;
+					break;
+				}
+
+				if (approach->name == NULL && approach->uniq == NULL)
+					break;
+
+				approach++;
+			}
+			if (device)
+				break;
+			bus++;
+		}
+		ret = fallback_or_device(db, device, builder->device_name, fallback);
+		if (ret) {
+			/* for multiple-match devices, set to the one we requested */
+			used_match = libwacom_match_new(used_match_name, used_match_uniq, *bus, vendor_id, product_id);
+			libwacom_set_default_match(ret, used_match);
+			libwacom_match_unref(used_match);
+		}
 	}
 
-	/* for multiple-match devices, set to the one we requested */
-	match = libwacom_match_new(match_name, bus, vendor_id, product_id);
-	libwacom_set_default_match(ret, match);
-	libwacom_match_unref(match);
-
-	/* if unset, use the kernel flags. Could be unset as well. */
-	if (device && ret->integration_flags == WACOM_DEVICE_INTEGRATED_UNSET)
-		ret->integration_flags = integration_flags;
-
-out:
-	g_free (name);
 	if (ret == NULL)
 		libwacom_error_set(error, WERROR_UNKNOWN_MODEL, "unknown model");
 	return ret;
 }
 
 LIBWACOM_EXPORT WacomDevice*
-libwacom_new_from_usbid(const WacomDeviceDatabase *db, int vendor_id, int product_id, WacomError *error)
+libwacom_new_from_path(const WacomDeviceDatabase *db, const char *path, WacomFallbackFlags fallback, WacomError *error)
 {
-	const WacomDevice *device;
+	int vendor_id, product_id;
+	WacomBusType bus;
+	WacomDevice *device;
+	WacomIntegrationFlags integration_flags;
+	char *name, *uniq;
+	WacomBuilder *builder;
 
-	if (!db) {
-		libwacom_error_set(error, WERROR_INVALID_DB, "db is NULL");
+	if (!path) {
+		libwacom_error_set(error, WERROR_INVALID_PATH, "path is NULL");
 		return NULL;
 	}
 
-	device = libwacom_new(db, NULL, vendor_id, product_id, WBUSTYPE_USB, error);
-	if (!device)
-		device = libwacom_new(db, NULL, vendor_id, product_id, WBUSTYPE_I2C, error);
-	if (!device)
-		device = libwacom_new(db, NULL, vendor_id, product_id, WBUSTYPE_BLUETOOTH, error);
+	if (!get_device_info (path, &vendor_id, &product_id, &name, &uniq, &bus, &integration_flags, error))
+		return NULL;
 
-	if (device)
-		return libwacom_copy(device);
+	builder = libwacom_builder_new();
+	libwacom_builder_set_match_name(builder, name);
+	libwacom_builder_set_bustype(builder, bus);
+	libwacom_builder_set_uniq(builder, uniq);
+	libwacom_builder_set_usbid(builder, vendor_id, product_id);
+	device = libwacom_new_from_builder(db, builder, fallback, error);
+	/* if unset, use the kernel flags. Could be unset as well. */
+	if (device && device->integration_flags == WACOM_DEVICE_INTEGRATED_UNSET)
+		device->integration_flags = integration_flags;
 
-	libwacom_error_set(error, WERROR_UNKNOWN_MODEL, NULL);
-	return NULL;
+	libwacom_builder_destroy(builder);
+	g_free (name);
+	g_free (uniq);
+
+	return device;
+}
+
+LIBWACOM_EXPORT WacomDevice*
+libwacom_new_from_usbid(const WacomDeviceDatabase *db, int vendor_id, int product_id, WacomError *error)
+{
+	WacomDevice *device;
+	WacomBuilder *builder = libwacom_builder_new();
+
+	libwacom_builder_set_usbid(builder, vendor_id, product_id);
+	device = libwacom_new_from_builder(db, builder, WFALLBACK_NONE, error);
+	libwacom_builder_destroy(builder);
+
+	return device;
 }
 
 LIBWACOM_EXPORT WacomDevice*
 libwacom_new_from_name(const WacomDeviceDatabase *db, const char *name, WacomError *error)
 {
-	const WacomDevice *device;
-	GList *keys, *l;
+	WacomBuilder *builder = libwacom_builder_new();
+	WacomDevice *device;
+	libwacom_builder_set_device_name(builder, name);
 
-	if (!db) {
-		libwacom_error_set(error, WERROR_INVALID_DB, "db is NULL");
-		return NULL;
-	}
+	device = libwacom_new_from_builder(db, builder, WFALLBACK_NONE, error);
+	libwacom_builder_destroy(builder);
 
-	g_return_val_if_fail(name != NULL, NULL);
-
-	device = NULL;
-	keys = g_hash_table_get_values (db->device_ht);
-	for (l = keys; l; l = l->next) {
-		WacomDevice *d = l->data;
-
-		if (g_str_equal(d->name, name)) {
-			device = d;
-			break;
-		}
-	}
-	g_list_free (keys);
-
-	if (device)
-		return libwacom_copy(device);
-
-	libwacom_error_set(error, WERROR_UNKNOWN_MODEL, NULL);
-	return NULL;
+	return device;
 }
 
 static void print_styli_for_device (int fd, const WacomDevice *device)
@@ -786,13 +975,18 @@ static void print_buttons_for_device (int fd, const WacomDevice *device)
 	print_button_flag_if(fd, device, "Bottom", WACOM_BUTTON_POSITION_BOTTOM);
 	print_button_flag_if(fd, device, "Touchstrip", WACOM_BUTTON_TOUCHSTRIP_MODESWITCH);
 	print_button_flag_if(fd, device, "Touchstrip2", WACOM_BUTTON_TOUCHSTRIP2_MODESWITCH);
+	print_button_flag_if(fd, device, "Dial", WACOM_BUTTON_DIAL_MODESWITCH);
 	print_button_flag_if(fd, device, "OLEDs", WACOM_BUTTON_OLED);
 	print_button_flag_if(fd, device, "Ring", WACOM_BUTTON_RING_MODESWITCH);
 	print_button_flag_if(fd, device, "Ring2", WACOM_BUTTON_RING2_MODESWITCH);
+	print_button_flag_if(fd, device, "Dial", WACOM_BUTTON_DIAL_MODESWITCH);
+	print_button_flag_if(fd, device, "Dial2", WACOM_BUTTON_DIAL2_MODESWITCH);
 	print_button_evdev_codes(fd, device);
 	dprintf(fd, "RingNumModes=%d\n", libwacom_get_ring_num_modes(device));
 	dprintf(fd, "Ring2NumModes=%d\n", libwacom_get_ring2_num_modes(device));
 	dprintf(fd, "StripsNumModes=%d\n", libwacom_get_strips_num_modes(device));
+	dprintf(fd, "DialNumModes=%d\n", libwacom_get_dial_num_modes(device));
+	dprintf(fd, "Dial2NumModes=%d\n", libwacom_get_dial2_num_modes(device));
 
 	dprintf(fd, "\n");
 }
@@ -817,6 +1011,7 @@ static void print_integrated_flags_for_device (int fd, const WacomDevice *device
 static void print_match(int fd, const WacomMatch *match)
 {
 	const char  *name       = libwacom_match_get_name(match);
+	const char  *uniq       = libwacom_match_get_uniq(match);
 	WacomBusType type	= libwacom_match_get_bustype(match);
 	int          vendor     = libwacom_match_get_vendor_id(match);
 	int          product    = libwacom_match_get_product_id(match);
@@ -830,9 +1025,11 @@ static void print_match(int fd, const WacomMatch *match)
 		case WBUSTYPE_UNKNOWN:		bus_name = "unknown";	break;
 		default:			g_assert_not_reached(); break;
 	}
-	dprintf(fd, "%s:%04x:%04x", bus_name, vendor, product);
+	dprintf(fd, "%s|%04x|%04x", bus_name, vendor, product);
 	if (name)
-		dprintf(fd, ":%s", name);
+		dprintf(fd, "|%s", name);
+	if (uniq)
+		dprintf(fd, "|%s", uniq);
 	dprintf(fd, ";");
 }
 
@@ -885,8 +1082,7 @@ libwacom_print_device_description(int fd, const WacomDevice *device)
 	dprintf(fd, "[Features]\n");
 	dprintf(fd, "Reversible=%s\n", libwacom_is_reversible(device)	? "true" : "false");
 	dprintf(fd, "Stylus=%s\n",	 libwacom_has_stylus(device)	? "true" : "false");
-	dprintf(fd, "Ring=%s\n",	 libwacom_has_ring(device)	? "true" : "false");
-	dprintf(fd, "Ring2=%s\n",	 libwacom_has_ring2(device)	? "true" : "false");
+	dprintf(fd, "NumRings=%s\n",	 libwacom_get_num_rings(device)	? "true" : "false");
 	dprintf(fd, "Touch=%s\n",	 libwacom_has_touch(device)	? "true" : "false");
 	dprintf(fd, "TouchSwitch=%s\n",	libwacom_has_touchswitch(device)? "true" : "false");
 	print_supported_leds(fd, device);
@@ -894,6 +1090,9 @@ libwacom_print_device_description(int fd, const WacomDevice *device)
 	dprintf(fd, "NumStrips=%d\n",	libwacom_get_num_strips(device));
 	dprintf(fd, "\n");
 
+	dprintf(fd, "NumDials=%d\n",	libwacom_get_num_dials(device));
+	dprintf(fd, "\n");
+
 	print_buttons_for_device(fd, device);
 }
 
@@ -957,13 +1156,16 @@ libwacom_match_unref(WacomMatch *match)
 
 	g_free (match->match);
 	g_free (match->name);
+	g_free (match->uniq);
 	g_free (match);
 
 	return NULL;
 }
 
 WacomMatch*
-libwacom_match_new(const char *name, WacomBusType bus, int vendor_id, int product_id)
+libwacom_match_new(const char *name, const char *uniq,
+		   WacomBusType bus,
+		   int vendor_id, int product_id)
 {
 	WacomMatch *match;
 	char *newmatch;
@@ -973,10 +1175,11 @@ libwacom_match_new(const char *name, WacomBusType bus, int vendor_id, int produc
 	if (name == NULL && bus == WBUSTYPE_UNKNOWN && vendor_id == 0 && product_id == 0)
 		newmatch = g_strdup("generic");
 	else
-		newmatch = make_match_string(name, bus, vendor_id, product_id);
+		newmatch = make_match_string(name, uniq, bus, vendor_id, product_id);
 
 	match->match = newmatch;
 	match->name = g_strdup(name);
+	match->uniq = g_strdup(uniq);
 	match->bus = bus;
 	match->vendor_id = vendor_id;
 	match->product_id = product_id;
@@ -984,6 +1187,56 @@ libwacom_match_new(const char *name, WacomBusType bus, int vendor_id, int produc
 	return match;
 }
 
+LIBWACOM_EXPORT WacomBuilder*
+libwacom_builder_new(void)
+{
+	WacomBuilder *builder = g_malloc0(sizeof(*builder));
+	return builder;
+}
+
+LIBWACOM_EXPORT void
+libwacom_builder_destroy(WacomBuilder *builder)
+{
+	g_free (builder->device_name);
+	g_free (builder->match_name);
+	g_free (builder->uniq);
+	g_free (builder);
+}
+
+LIBWACOM_EXPORT void
+libwacom_builder_set_bustype(WacomBuilder *builder, WacomBusType bustype)
+{
+	builder->bus = bustype;
+}
+
+LIBWACOM_EXPORT void
+libwacom_builder_set_usbid(WacomBuilder *builder, int vendor_id, int product_id)
+{
+	builder->vendor_id = vendor_id;
+	builder->product_id = product_id;
+}
+
+LIBWACOM_EXPORT void
+libwacom_builder_set_device_name(WacomBuilder *builder, const char *name)
+{
+	g_free(builder->device_name);
+	builder->device_name = g_strdup(name);
+}
+
+LIBWACOM_EXPORT void
+libwacom_builder_set_match_name(WacomBuilder *builder, const char *name)
+{
+	g_free(builder->match_name);
+	builder->match_name = g_strdup(name);
+}
+
+LIBWACOM_EXPORT void
+libwacom_builder_set_uniq(WacomBuilder *builder, const char *uniq)
+{
+	g_free(builder->uniq);
+	builder->uniq = g_strdup(uniq);
+}
+
 void
 libwacom_add_match(WacomDevice *device, WacomMatch *newmatch)
 {
@@ -1142,13 +1395,19 @@ libwacom_get_supported_styli(const WacomDevice *device, int *num_styli)
 LIBWACOM_EXPORT int
 libwacom_has_ring(const WacomDevice *device)
 {
-	return !!(device->features & FEATURE_RING);
+	return device->num_rings >= 1;
 }
 
 LIBWACOM_EXPORT int
 libwacom_has_ring2(const WacomDevice *device)
 {
-	return !!(device->features & FEATURE_RING2);
+	return device->num_rings >= 2;
+}
+
+LIBWACOM_EXPORT int
+libwacom_get_num_rings(const WacomDevice *device)
+{
+	return device->num_rings;
 }
 
 LIBWACOM_EXPORT int
@@ -1175,6 +1434,24 @@ libwacom_get_strips_num_modes(const WacomDevice *device)
 	return device->strips_num_modes;
 }
 
+LIBWACOM_EXPORT int
+libwacom_get_num_dials(const WacomDevice *device)
+{
+	return device->num_dials;
+}
+
+LIBWACOM_EXPORT int
+libwacom_get_dial_num_modes(const WacomDevice *device)
+{
+	return device->dial_num_modes;
+}
+
+LIBWACOM_EXPORT int
+libwacom_get_dial2_num_modes(const WacomDevice *device)
+{
+	return device->dial2_num_modes;
+}
+
 LIBWACOM_EXPORT const WacomStatusLEDs *
 libwacom_get_status_leds(const WacomDevice *device, int *num_leds)
 {
@@ -1441,6 +1718,12 @@ libwacom_match_get_name(const WacomMatch *match)
 	return match->name;
 }
 
+LIBWACOM_EXPORT const char *
+libwacom_match_get_uniq(const WacomMatch *match)
+{
+	return match->uniq;
+}
+
 LIBWACOM_EXPORT WacomBusType
 libwacom_match_get_bustype(const WacomMatch *match)
 {
